@@ -1,6 +1,7 @@
 import asyncio
 import inspect
 import logging
+from datetime import datetime, timedelta
 from functools import wraps
 from os import getenv
 from pathlib import Path
@@ -14,7 +15,7 @@ from sqlalchemy.sql import text
 
 from spellbot._version import __version__
 from spellbot.assets import ASSET_FILES, s
-from spellbot.data import AuthorizedChannel, BotPrefix, Data, Game, Tag, User
+from spellbot.data import AuthorizedChannel, Data, Game, Server, Tag, User, WaitTime
 
 # Application Paths
 RUNTIME_ROOT = Path(".")
@@ -27,8 +28,14 @@ MIGRATIONS_DIR = SCRIPTS_DIR / "migrations"
 # Application Settings
 ADMIN_ROLE = "SpellBot Admin"
 CREATE_ENDPOINT = "https://us-central1-magic-night-30324.cloudfunctions.net/createGame"
-CHECK_INTERVAL_SEC = 30
-EXPIRE_WINDOW_MIN = 30
+AVG_QUEUE_TIME_WINDOW_MIN = 30
+
+
+def to_int(s):
+    try:
+        return int(s)
+    except ValueError:
+        return None
 
 
 def is_admin(channel, user_or_member):
@@ -47,6 +54,33 @@ def ensure_application_directories_exist():
     DB_DIR.mkdir(exist_ok=True)
 
 
+def paginate(text):
+    """Discord responses must be 2000 characters of less; paginate breaks them up."""
+    breakpoints = ["\n", ".", ",", "-"]
+    remaining = text
+    while len(remaining) > 2000:
+        breakpoint = 1999
+
+        for char in breakpoints:
+            index = remaining.rfind(char, 1800, 1999)
+            if index != -1:
+                breakpoint = index
+                break
+
+        message = remaining[0 : breakpoint + 1]
+        yield message
+        remaining = remaining[breakpoint + 1 :]
+        last_line_end = message.rfind("\n")
+        if last_line_end != -1 and len(message) > last_line_end + 1:
+            last_line_start = last_line_end + 1
+        else:
+            last_line_start = 0
+        if message[last_line_start] == ">":
+            remaining = f"> {remaining}"
+
+    yield remaining
+
+
 def command(allow_dm=True):
     """Decorator for bot command methods."""
 
@@ -91,29 +125,34 @@ class SpellBot(discord.Client):
             if hasattr(member[1], "is_command") and member[1].is_command
         ]
 
-        # Start up our periodic task to clean up expired games now.
+        self._begin_background_tasks(loop)
+
+    def _begin_background_tasks(self, loop):  # pragma: no cover
+        """Start up any periodic background tasks."""
         self.cleanup_expired_games_task(loop)
+        self.cleanup_expired_wait_times_task(loop)
 
     def cleanup_expired_games_task(self, loop):  # pragma: no cover
         """Starts a task that deletes old games."""
+        THIRTY_SECONDS = 30
 
         async def task():
             while True:
-                await self.cleanup_expired_games(EXPIRE_WINDOW_MIN)
-                await asyncio.sleep(CHECK_INTERVAL_SEC)
+                await self.cleanup_expired_games()
+                await asyncio.sleep(THIRTY_SECONDS)
 
         loop.create_task(task())
 
-    async def cleanup_expired_games(self, window):
+    async def cleanup_expired_games(self):
         """Deletes games older than the given window of minutes."""
         session = self.data.Session()
         try:
-            expired = Game.expired(session, window)
+            expired = Game.expired(session)
             for game in expired:
                 for user in game.users:
                     discord_user = self.get_user(user.xid)
                     if discord_user:
-                        await discord_user.send(s("expired", window=window))
+                        await discord_user.send(s("expired", window=game.server.expire))
                 session.delete(game)
             session.commit()
         except exc.SQLAlchemyError as e:
@@ -123,6 +162,31 @@ class SpellBot(discord.Client):
         finally:
             session.close()
 
+    def cleanup_expired_wait_times_task(self, loop):  # pragma: no cover
+        """Starts a task that deletes old wait times data."""
+        FIVE_MINUTES = 300
+
+        async def task():
+            while True:
+                await self.cleanup_expired_waits(AVG_QUEUE_TIME_WINDOW_MIN)
+                await asyncio.sleep(FIVE_MINUTES)
+
+        loop.create_task(task())
+
+    async def cleanup_expired_waits(self, window):
+        """Deletes wait time data older than the given window of minutes."""
+        session = self.data.Session()
+        try:
+            cutoff = datetime.utcnow() - timedelta(minutes=window)
+            session.query(WaitTime).filter(WaitTime.created_at < cutoff).delete()
+            session.commit()
+        except exc.SQLAlchemyError as e:
+            logging.error("error: cleanup_expired_waits:", e)
+            session.rollback()
+            raise
+        finally:
+            session.close()
+
     def run(self):  # pragma: no cover
         super().run(self.token)
 
@@ -139,6 +203,14 @@ class SpellBot(discord.Client):
             self.session.add(db_user)
         return db_user
 
+    def ensure_server_exists(self, guild_xid):
+        """Ensures that the server row exists for the given discord guild id."""
+        server = self.session.query(Server).filter(Server.guild_xid == guild_xid).first()
+        if not server:
+            server = Server(guild_xid=guild_xid)
+            self.session.add(server)
+        return server
+
     @property
     def commands(self):
         """Returns a list of commands supported by this bot."""
@@ -207,7 +279,7 @@ class SpellBot(discord.Client):
         # only respond to command-like messages
         if not private:
             rows = self.data.conn.execute(
-                text("SELECT prefix FROM bot_prefixes WHERE guild_xid = :g"),
+                text("SELECT prefix FROM servers WHERE guild_xid = :g"),
                 g=message.channel.guild.id,
             )
             prefixes = [row.prefix for row in rows]
@@ -261,7 +333,7 @@ class SpellBot(discord.Client):
         """
         Sends you this help message.
         """
-        usage = "__**SpellBot Usage**__\n"
+        usage = ""
         for command in self.commands:
             method = getattr(self, command)
             doc = method.__doc__.split("&")
@@ -269,14 +341,24 @@ class SpellBot(discord.Client):
             use = inspect.cleandoc(use)
             use = use.replace("\n", "\n> ")
 
-            title = f"**{prefix}{command}**"
+            title = f"{prefix}{command}"
             if params:
-                title = f"{title} _{params}_"
-            usage += f"\n{title}"
+                title = f"{title} {params}"
+            usage += f"\n`{title}`"
             usage += f"\n>  {use}"
             usage += "\n"
-        usage += "\n_SpellBot created by amy@lexicalunit.com_"
-        await author.send(usage, file=None)
+        usage += "---"
+        usage += (
+            " \nPlease report any bugs and suggestions at"
+            " <https://github.com/lexicalunit/spellbot/issues>!"
+        )
+        usage += "\n"
+        usage += (
+            "\n💜 You can help keep SpellBot running by supporting me on Ko-fi! "
+            "<https://ko-fi.com/Y8Y51VTHZ>"
+        )
+        for page in paginate(usage):
+            await author.send(page)
 
     @command(allow_dm=True)
     async def about(self, prefix, channel, author, mentions, params):
@@ -296,11 +378,14 @@ class SpellBot(discord.Client):
         author = "[@lexicalunit](https://github.com/lexicalunit)"
         embed.add_field(name="Author", value=author)
         embed.description = (
-            "_A Discord bot for SpellTable._\n"
+            "_A Discord bot for [SpellTable](https://www.spelltable.com/)._\n"
             "\n"
             f"Use the command `{prefix}help` for usage details. "
             "Having issues with SpellBot? "
             "Please [report bugs](https://github.com/lexicalunit/spellbot/issues)!\n"
+            "\n"
+            "💜 Help keep SpellBot running by "
+            "[supporting me on Ko-fi!](https://ko-fi.com/Y8Y51VTHZ)"
         )
         embed.url = "https://github.com/lexicalunit/spellbot"
         embed.set_footer(text="MIT © amy@lexicalunit et al")
@@ -308,48 +393,58 @@ class SpellBot(discord.Client):
         await channel.send(embed=embed, file=None)
 
     @command(allow_dm=False)
-    async def queue(self, prefix, channel, author, mentions, params):
+    async def play(self, prefix, channel, author, mentions, params):
         """
-        Enter the queue for a game on SpellTable.
+        Enter a play queue for a game on SpellTable.
 
-        You can get in the queue with a friend by including them in the command with as
-        many @ mentions as you want. You can also choose the number of players by using
-        `size:2` for a two player game, for example. The default number of players is 4.
+        You can get in a queue with a friend by mentioning them in the command with the @
+        character. You can also change the number of players from the default of four by
+        using, for example, `!play size:2` to create a two player game.
 
-        You can also give a list of tags and you will only be matched against users who
-        have the same tags as you. For example, if you want to enter into a cEDH queue
-        you could try `queue cEDH`. Check with your server for supported tags. A valid
-        tag can not be just a number, like `10` for example, nor can it be longer than
-        50 characters.
-        & [@mention-1] [@mention-2] [...] [size:N] [tag-1] [tag-2] [...]
+        Up to five tags can be given as well. For example, `!play no-combo proxy` has
+        two tags: `no-combo` and `proxy`. Look on your server for what tags are being
+        used by your community members. Tags can **not** be a number like `5`. Be careful
+        when using tags because the matchmaker will only pair you up with other players
+        who've used **EXACTLY** the same tags.
+
+        You can also specify a power level like `!play power:7` for example and the
+        matchmaker will attempt to find a game with similar power levels for you. Note
+        that players who specify a power level will never get paired up with players who
+        have not, and vice versa. You will also not be matched up _exactly_ by power level
+        as there is a fudge factor involved.
+
+        If your server's admins have set the scope for your server to "channel", then
+        matchmaking will only happen between other players who run this command in the
+        same channel as you did. The default scope for matchmaking is server-wide.
+        & [@mention-1] [@mention-2] [...] [size:N] [power:N] [tag-1] [tag-2] [...]
         """
         params = [param.lower() for param in params]
         user = self.ensure_user_exists(author)
         if user.waiting:
-            return await author.send(s("queue_already"))
-
-        def to_int(s):
-            try:
-                return int(s)
-            except ValueError:
-                return None
+            return await author.send(s("play_already"))
 
         size = 4
+        power = None
         for param in params:
             if param.startswith("size:"):
                 size = to_int(param.replace("size:", ""))
+            elif param.startswith("power:"):
+                power = to_int(param.replace("power:", ""))
 
         if not size or not (1 < size < 5):
-            return await author.send(s("queue_size_bad"))
+            return await author.send(s("play_size_bad"))
+
+        if power and not (1 <= power <= 10):
+            return await author.send(s("play_power_bad"))
 
         if len(mentions) >= size:
-            return await author.send(s("queue_too_many"))
+            return await author.send(s("play_too_many_mentions"))
 
         mentioned_users = []
         for mentioned in mentions:
             mentioned_user = self.ensure_user_exists(mentioned)
             if mentioned_user.waiting:
-                return await author.send(s("queue_mention_already", user=mentioned))
+                return await author.send(s("play_mention_already", user=mentioned))
             mentioned_users.append(mentioned_user)
 
         tags = []
@@ -357,12 +452,16 @@ class SpellBot(discord.Client):
             param
             for param in params
             if not param.startswith("size:")
+            and not param.startswith("power:")
+            and not param.startswith("<")
             and not param.startswith("@")
             and not param.isdigit()
             and not len(param) >= 50
         ]
         if not tag_names:
             tag_names = ["default"]
+        if len(tag_names) > 5:
+            return await author.send(s("play_too_many_tags"))
         for tag_name in tag_names:
             tag = self.session.query(Tag).filter_by(name=tag_name).first()
             if not tag:
@@ -370,23 +469,70 @@ class SpellBot(discord.Client):
                 self.session.add(tag)
             tags.append(tag)
 
+        server = self.ensure_server_exists(channel.guild.id)
         user.enqueue(
-            size=size, guild_xid=channel.guild.id, include=mentioned_users, tags=tags
+            server=server,
+            channel_xid=channel.id,
+            include=mentioned_users,
+            size=size,
+            power=power,
+            tags=tags,
         )
         self.session.commit()
 
+        found_discord_users = []
         if len(user.game.users) == size:
+            for game_user in user.game.users:
+                discord_user = self.get_user(game_user.xid)
+                if not discord_user:  # game_user has left the server since queueing
+                    game_user.dequeue()
+                else:
+                    found_discord_users.append(discord_user)
+
+        if len(found_discord_users) == size:  # all players matched, game is ready
             game_url = self.create_game()
-            for player in user.game.users:
-                discord_user = self.get_user(player.xid)
-                if discord_user:
-                    await discord_user.send(s("queue_ready", url=game_url))
+            players_s = ", ".join(
+                [f"<@!{discord_user.id}>" for discord_user in found_discord_users]
+            )
+            tags_s = ", ".join([tag.name for tag in user.game.tags])
+            channel_s = f"<#{channel.id}>"
+            power_s = str(user.game.power)
+            response = s(
+                "play_ready",
+                url=game_url,
+                players=players_s,
+                tags=tags_s,
+                channel=channel_s,
+                power=power_s,
+            )
+
+            for game_user, discord_user in zip(user.game.users, found_discord_users):
+                await discord_user.send(response)
+                dequeue_at = datetime.utcnow()
+                seconds = (dequeue_at - game_user.queued_at).total_seconds()
+                WaitTime.log(
+                    self.session,
+                    guild_xid=server.guild_xid,
+                    channel_xid=channel.id,
+                    seconds=seconds,
+                )
             self.session.delete(user.game)
-        else:
+        else:  # still waiting on more players, game is pending
+            average = WaitTime.average(
+                self.session,
+                guild_xid=server.guild_xid,
+                channel_xid=channel.id,
+                scope=server.scope,
+                window_min=AVG_QUEUE_TIME_WINDOW_MIN,
+            )
             for player in user.game.users:
                 discord_user = self.get_user(player.xid)
                 if discord_user:
-                    await discord_user.send(s("queue"))
+                    if average:
+                        response = s("play_queue_with_average", average=f"{average:.2f}")
+                        await discord_user.send(response)
+                    else:
+                        await discord_user.send(s("play_queue"))
 
     @command(allow_dm=True)
     async def leave(self, prefix, channel, author, mentions, params):
@@ -400,30 +546,55 @@ class SpellBot(discord.Client):
         user.dequeue()
         await author.send(s("leave"))
 
+    @command(allow_dm=False)
+    async def status(self, prefix, channel, author, mentions, params):
+        """
+        Show some details about the queues on your server.
+        """
+        server = self.ensure_server_exists(channel.guild.id)
+        average = WaitTime.average(
+            self.session,
+            guild_xid=server.guild_xid,
+            channel_xid=channel.id,
+            scope=server.scope,
+            window_min=AVG_QUEUE_TIME_WINDOW_MIN,
+        )
+        if average:
+            await channel.send(s("status", average=f"{average:.2f}"))
+        else:
+            await channel.send(s("status_unknown"))
+
     @command(allow_dm=False)
     async def spellbot(self, prefix, channel, author, mentions, params):
         """
-        Configure SpellBot for your server. Use with one of the following subcommands:
+        Configure SpellBot for your server.
 
-        - channel <list>: Set SpellBot to only respond in the given list of channels
-        - prefix <string>: Set SpellBot prefix for commands in text channels
+        The following subcommands are supported:
+
+        * `channel <list>`: Set SpellBot to only respond in the given list of channels.
+        * `prefix <string>`: Set SpellBot prefix for commands in text channels.
+        * `scope <server|channel>`: Set matchmaking scope to server-wide or channel-only.
+        * `exxpire <number>`: Set the number of minutes before pending games expire.
+
+        _You must have the "SpellBot Admin" role to use any of these commands._
         & <subcommand> [subcommand parameters]
         """
         if not is_admin(channel, author):
             return await author.send(s("not_admin"))
         if not params:
             return await author.send(s("spellbot_missing_subcommand"))
+        self.ensure_server_exists(channel.guild.id)
         command = params[0]
         if command == "channels":
-            return await self.spellbot_channels(
-                prefix, channel, author, mentions, params[1:]
-            )
+            await self.spellbot_channels(prefix, channel, author, mentions, params[1:])
         elif command == "prefix":
-            return await self.spellbot_prefix(
-                prefix, channel, author, mentions, params[1:]
-            )
+            await self.spellbot_prefix(prefix, channel, author, mentions, params[1:])
+        elif command == "scope":
+            await self.spellbot_scope(prefix, channel, author, mentions, params[1:])
+        elif command == "expire":
+            await self.spellbot_expire(prefix, channel, author, mentions, params[1:])
         else:
-            return await author.send(s("spellbot_unknown_subcommand", command=command))
+            await author.send(s("spellbot_unknown_subcommand", command=command))
 
     async def spellbot_channels(self, prefix, channel, author, mentions, params):
         if not params:
@@ -441,31 +612,68 @@ class SpellBot(discord.Client):
         if not params:
             return await author.send(s("spellbot_prefix_none"))
         prefix_str = params[0][0:10]
-        prefix = (
-            self.session.query(BotPrefix)
-            .filter(BotPrefix.guild_xid == channel.guild.id)
-            .first()
+        server = (
+            self.session.query(Server)
+            .filter(Server.guild_xid == channel.guild.id)
+            .one_or_none()
         )
-        if prefix:
-            prefix.prefix = prefix_str
+        if server:
+            server.prefix = prefix_str
         else:
-            self.session.add(BotPrefix(guild_xid=channel.guild.id, prefix=prefix_str))
+            self.session.add(Server(guild_xid=channel.guild.id, prefix=prefix_str))
         return await channel.send(s("spellbot_prefix", prefix=prefix_str))
 
+    async def spellbot_scope(self, prefix, channel, author, mentions, params):
+        if not params:
+            return await author.send(s("spellbot_scope_none"))
+        scope_str = params[0].lower()
+        if scope_str not in ("server", "channel"):
+            return await author.send(s("spellbot_scope_bad"))
+        server = (
+            self.session.query(Server)
+            .filter(Server.guild_xid == channel.guild.id)
+            .one_or_none()
+        )
+        if server:
+            server.scope = scope_str
+        else:
+            self.session.add(Server(guild_xid=channel.guild.id, scope=scope_str))
+        return await channel.send(s("spellbot_scope", scope=scope_str))
+
+    async def spellbot_expire(self, prefix, channel, author, mentions, params):
+        if not params:
+            return await author.send(s("spellbot_expire_none"))
+        expire = to_int(params[0])
+        if not expire or not (10 < expire < 60):
+            return await author.send(s("spellbot_expire_bad"))
+        server = (
+            self.session.query(Server)
+            .filter(Server.guild_xid == channel.guild.id)
+            .one_or_none()
+        )
+        if server:
+            server.expire = expire
+        else:
+            self.session.add(Server(guild_xid=channel.guild.id, expire=expire))
+        return await author.send(s("spellbot_expire", expire=expire))
+
 
 def get_db_env(fallback):  # pragma: no cover
     """Returns the database env var from the environment or else the given gallback."""
-    return getenv("SPELLTABLE_DB_ENV", fallback)
+    value = getenv("SPELLTABLE_DB_ENV", fallback)
+    return value or fallback
 
 
 def get_db_url(database_env, fallback):  # pragma: no cover
     """Returns the database url from the environment or else the given fallback."""
-    return getenv(database_env, fallback)
+    value = getenv(database_env, fallback)
+    return value or fallback
 
 
 def get_log_level(fallback):  # pragma: no cover
     """Returns the log level from the environment or else the given gallback."""
-    return getenv("SPELLTABLE_LOG_LEVEL", fallback)
+    value = getenv("SPELLTABLE_LOG_LEVEL", fallback)
+    return value or fallback
 
 
 @click.command()
